"""Rough Mount Fuji (RMF) fitness landscapes w/ tunable ruggedness using Numpy.
See J Neidhart, IG Szendro, J Krug 
    Adaptation in Tunably Rugged Fitness Landscapes: The Rough Mount Fuji Model.
    Genetics 2014 .
    DOI: https://doi.org/10.1534/genetics.114.167668 
See Aita et al. 
    Analysis of a local fitness landscape with a model of the rough Mt. Fuji-type landscape: Application to prolyl endopeptidase and thermolysin. 
    Biopolymers 2000 .
    DOI: https://doi.org/10.1002/(SICI)1097-0282(200007)54:1<64::AID-BIP70>3.0.CO;2-R
"""
